@@ -1,5 +1,12 @@
 /**********************************************************************
+atom.cpp - Handle OBAtom class
+
 Copyright (C) 1998-2001 by OpenEye Scientific Software, Inc.
+Some portions Copyright (c) 2001-2003 by Geoffrey R. Hutchison
+Some portions Copyright (c) 2003 by Michael Banck
+
+This file is part of the Open Babel project.
+For more information, see <http://openbabel.sourceforge.net/>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -16,8 +23,75 @@ GNU General Public License for more details.
 #include "molchrg.h"
 #include "phmodel.h"
 
+#include "math/matrix3x3.h"
+
+using namespace std;
+
 namespace OpenBabel {
 
+/** \class OBAtom
+    \brief Atom class
+
+ To understand the OBAtom class it is important to state a key
+ decision on which the design was based. In OBabel the atom class
+ existed, but it was only a data container. All data access and
+ modification of atoms was done through the molecule. The result
+ was a molecule class that was very large an unwieldy. So the OBAtom
+ class was made smarter, and many of the atom-specific routines
+ were separated into the OBAtom thereby decentralizing and
+ shrinking the OBMol class. As a result the OBAtom class not only
+ holds data, but facilitates extraction of data perceived from both
+ the atom and the molecule.
+
+ A number of data extraction methods perform what is called
+ `Lazy Evaluation,' which is essentially on-the-fly evaluation.
+ For example, when an atom is queried as to whether it is cyclic
+ or what it's hybridization state is the information is perceived
+ automatically. The perception of a particular trait is actually
+ performed on the entire molecule the first time it is requested of
+ an atom or bond, and stored for subsequent requests for the same
+ trait of additional atoms or bonds.The OBAtom class is similar to
+ OBMol and the whole of Open Babel in that data access and modification
+ is done through Get and Set methods.
+
+ The following code demonstrates how to print out the atom numbers,
+ element numbers, and coordinates of a molecule:
+\code
+   OBMol mol(SDF,SDF);
+   OBFileFormat::ReadMolecule(cin, mol);
+   OBAtom *atom;
+   vector<OBNodeBase*>::iterator i;
+   for (atom = mol.BeginAtom(i);atom;atom = mol.NextAtom(i))
+   {
+       cout << atom->GetIdx() << ` `;
+       cout << atom->GetAtomicNum() << ` `;
+       cout << atom->GetVector() << endl;
+   }
+\endcode
+ A number of the property member functions indicate that atoms
+ have some knowlege of their covalently attached neighbor atoms.
+ Bonding information is partly redundant within a molecule in
+ that an OBMol has a complete list of bonds in a molecule, and
+ an OBAtom has a list bonds of which it is a member. The following
+ code demonstrates how an OBAtom uses its bond information to loop
+ over atoms attached to itself:
+\code
+   OBMol mol(SDF,SDF);
+   OBFileFormat::ReadMolecule(cin, mol);
+   OBAtom *atom,*nbr;
+   vector<OBEdgeBase*>::iterator i;
+   atom = mol.GetAtom(1);
+   for (nbr = atom->BeginNbrAtom(i);nbr;nbr = atom->NextNbrAtom(i))
+   {
+   cout << "atom #" << atom->GetIdx() << " is attached to atom #" << nbr->GetIdx() << endl;
+   }
+\endcode
+ should produce an output like
+\code
+   atom #1 is attached to atom #2
+\endcode
+*/
+
 extern OBAromaticTyper  aromtyper;
 extern OBAtomTyper      atomtyper;
 extern OBPhModel        phmodel;
@@ -28,56 +102,52 @@ extern OBPhModel        phmodel;
 
 OBAtom::OBAtom()
 {
-  _c = (float**)NULL;
   _parent = (OBMol*)NULL;
-  _cidx = 0;
-  _flags=0;
-  _idx = 0;
-  _hyb = 0;
-  _ele = (char)0;
-  _impval = 0;
-  _fcharge = 0;
-  //_stereo = 0;
-  _type[0] = '\0';
-  _pcharge = 0.0;
-  _vbond.clear();
-  _vbond.reserve(4);
-  _residue = (OBResidue*)NULL;
+  Clear();
 }
 
 OBAtom::~OBAtom()
 {
     if (_residue != NULL)
         _residue->RemoveAtom(this);
+    if (!_vdata.empty())
+      {
+	vector<OBGenericData*>::iterator m;
+	for (m = _vdata.begin();m != _vdata.end();m++) delete *m;
+	_vdata.clear();
+      }
 }
 
 void OBAtom::Clear()
 {
-  _c = (float**)NULL;
+  _c = (double**)NULL;
   _cidx = 0;
   _flags=0;
   _idx = 0;
   _hyb = 0;
   _ele = (char)0;
+  _isotope = 0;
+  _spinmultiplicity=0; // CM 18 Sept 2003
   _impval = 0;
   _fcharge = 0;
-  //_stereo = 0;
   _type[0] = '\0';
   _pcharge = 0.0;
   _vbond.clear();
   _vbond.reserve(4);
   _residue = (OBResidue*)NULL;
+  _vdata.clear();
 }
 
 OBAtom &OBAtom::operator=(OBAtom &src)
      //copy atom information 
      //bond info is not copied here as ptrs may be invalid
+     //vdata is also not copied yet (again it's unclear what can work)
 {
   _idx = src.GetIdx();
   _hyb = src.GetHyb();
   _ele = src.GetAtomicNum();
+  _isotope = src.GetIsotope();
   _fcharge = src.GetFormalCharge();
-  //_stereo = src.GetStereo();
   strcpy(_type,src.GetType());
   _pcharge = src.GetPartialCharge();
   _v = src.GetVector();
@@ -132,7 +202,7 @@ bool OBAtom::IsOneFour(OBAtom *a1)
 
 bool OBAtom::IsAxial()
 {
-  float tor;
+  double tor;
   OBAtom *a,*b,*c;
   vector<OBEdgeBase*>::iterator i,j,k;
   
@@ -188,6 +258,17 @@ int OBAtom::CountBondsOfOrder(unsigned int order)
   return(count);
 }
 
+bool OBAtom::HasNonSingleBond()
+{
+  OBBond *bond;
+  vector<OBEdgeBase*>::iterator i;
+  for (bond = BeginBond(i);bond;bond = NextBond(i))
+    if (bond->GetBO() != 1)
+      return(true);
+
+  return(false);
+}
+
 bool OBAtom::IsPolarHydrogen()
 {
   if (!IsHydrogen()) return(false);
@@ -223,7 +304,7 @@ bool OBAtom::IsNonPolarHydrogen()
   return(false);
 }
 
-Vector &OBAtom::GetVector()
+vector3 &OBAtom::GetVector()
 {
   if (!_c) return(_v);
 
@@ -237,7 +318,7 @@ void OBAtom::SetVector()
   if (_c) _v.Set((*_c)[_cidx],(*_c)[_cidx+1],(*_c)[_cidx+2]);
 }
 
-void OBAtom::SetVector(Vector &v)
+void OBAtom::SetVector(vector3 &v)
 {
   if (!_c) _v = v;
   else
@@ -248,7 +329,7 @@ void OBAtom::SetVector(Vector &v)
     }
 }
 
-void OBAtom::SetVector(const float x,const float y,const float z)
+void OBAtom::SetVector(const double x,const double y,const double z)
 {
   if (!_c) _v.Set(x,y,z);
   else
@@ -259,6 +340,30 @@ void OBAtom::SetVector(const float x,const float y,const float z)
     }
 }
 
+void OBAtom::SetType(char *type)
+{
+  strcpy(_type,type);
+  if (_ele == 1 && type[0] == 'D')
+    _isotope = 2;
+}
+
+void OBAtom::SetType(string &type)
+{
+  strcpy(_type,type.c_str());
+  if (_ele == 1 && type[0] == 'D')
+    _isotope = 2;
+}
+
+void OBAtom::SetIsotope(unsigned int iso)
+{
+  if (_ele == 1 && iso == 2)
+    SetType("D");
+  else if (_ele == 1 && (iso == 1 || iso == 0))
+    SetType("H");
+
+  _isotope = iso;
+}
+
 OBAtom *OBAtom::GetNextAtom()
 {
   OBMol *mol = (OBMol*)GetParent();
@@ -272,18 +377,44 @@ OBResidue *OBAtom::GetResidue()
     else if (!((OBMol*)GetParent())->HasChainsPerceived())
     {
         chainsparser.PerceiveChains(*((OBMol*)GetParent()));
+        ((OBMol*)GetParent())->SetChainsPerceived();
         return _residue;
     }
     else
         return NULL;
 }
 
+double OBAtom::GetAtomicMass() const
+{
+  if (_isotope == 0)
+    return etab.GetMass(_ele);
+  else
+    return isotab.GetExactMass(_ele, _isotope);
+}
+
+double OBAtom::GetExactMass() const
+{
+  return isotab.GetExactMass(_ele, _isotope);
+}
+
 char *OBAtom::GetType()
 {
   OBMol *mol = (OBMol*)GetParent();
   if (mol && !mol->HasAtomTypesPerceived())
     atomtyper.AssignTypes(*((OBMol*)GetParent()));
 
+  if (strlen(_type) == 0) // Somehow we still don't have a type!
+    {
+      char num[6];
+      OBTypeTable tempTable;
+      tempTable.SetFromType("ATN");
+      tempTable.SetToType("INT");
+      snprintf(num, 6, "%d", GetAtomicNum());
+      tempTable.Translate(_type, num);
+    }
+  if (_ele == 1 && _isotope == 2)
+    snprintf(_type, 6, "%s", "D");
+
   return(_type);
 }
 
@@ -308,21 +439,19 @@ unsigned int OBAtom::GetHyb() const
 
 
 unsigned int OBAtom::GetHvyValence() const
-     //returns the number of non-hydrogens connected to an atom
 {
   unsigned int count=0;
 
-  OBAtom *atom;
+  OBBond *bond;
   vector<OBEdgeBase*>::iterator i;
-  for (atom = ((OBAtom*)this)->BeginNbrAtom(i);atom;atom = ((OBAtom*)this)->NextNbrAtom(i))
-	  if (!atom->IsHydrogen())
-		  count++;
+  for (bond = ((OBAtom*)this)->BeginBond(i); bond; bond = ((OBAtom*)this)->NextBond(i))
+    if (!(bond->GetNbrAtom((OBAtom*)this)->IsHydrogen()))
+	count++;
 
   return(count);
 }
 
 unsigned int OBAtom::GetHeteroValence() const
-     //returns the number of heteroatoms connected to an atom
 {
   unsigned int count=0;
   OBBond *bond;
@@ -334,7 +463,7 @@ unsigned int OBAtom::GetHeteroValence() const
   return((unsigned int)count);
 }
 
-float OBAtom::GetPartialCharge()
+double OBAtom::GetPartialCharge()
 {
   if (!GetParent()) return(_pcharge);
   if (!((OBMol*)GetParent())->AutomaticPartialCharge()) return(_pcharge);
@@ -342,6 +471,12 @@ float OBAtom::GetPartialCharge()
   if (!((OBMol*)GetParent())->HasPartialChargesPerceived())
     {
       //seed partial charges are set in the atom typing procedure
+      OBAtom *atom;
+      OBMol *mol = (OBMol*)GetParent();
+      vector<OBNodeBase*>::iterator i;
+      for (atom = mol->BeginAtom(i); atom; atom = mol->NextAtom(i))
+	atom->SetPartialCharge(0.0);
+
       phmodel.AssignSeedPartialCharge(*((OBMol*)GetParent()));
       OBGastChrg gc;
       gc.AssignPartialCharges(*((OBMol*)GetParent()));
@@ -485,19 +620,19 @@ bool OBAtom::IsNitroOxygen()
 
 bool OBAtom::IsHeteroatom()
 {
-	switch(GetAtomicNum())
-	{
-	case 7:
-	case 8:
-	case 9:
-	case 15:
-	case 16:
-	case 17:
-	case 35:
-	case 53:
-		return(true);
-	}
-	return(false);
+  switch(GetAtomicNum())
+    {
+    case 7:
+    case 8:
+    case 9:
+    case 15:
+    case 16:
+    case 17:
+    case 35:
+    case 53:
+      return(true);
+    }
+  return(false);
 }
 
 bool OBAtom::IsAromatic() const
@@ -565,6 +700,7 @@ unsigned int OBAtom::MemberOfRingCount() const
 {
   vector<OBRing*> rlist;
   vector<OBRing*>::iterator i;
+  unsigned int count=0;
 
   OBMol *mol = (OBMol*)((OBAtom*)this)->GetParent();
 
@@ -573,7 +709,6 @@ unsigned int OBAtom::MemberOfRingCount() const
 
   if (!((OBAtom*)this)->IsInRing()) return(0);
 
-  int count=0;
   rlist = mol->GetSSSR();
 
   for (i = rlist.begin();i != rlist.end();i++)
@@ -583,15 +718,65 @@ unsigned int OBAtom::MemberOfRingCount() const
   return((unsigned int)count);
 }
 
+unsigned int OBAtom::MemberOfRingSize() const
+{
+  vector<OBRing*> rlist;
+  vector<OBRing*>::iterator i;
+
+  OBMol *mol = (OBMol*)((OBAtom*)this)->GetParent();
+
+  if (!mol->HasSSSRPerceived())
+    mol->FindSSSR();
+
+  if (!((OBAtom*)this)->IsInRing()) return(0);
+
+  rlist = mol->GetSSSR();
+
+  for (i = rlist.begin();i != rlist.end();i++)
+    if ((*i)->IsInRing(GetIdx()))
+      return((*i)->Size());
+  
+  return(0);
+}
+
+double	  OBAtom::SmallestBondAngle()
+{
+  OBAtom *b, *c;
+  vector3 v1, v2;
+  double degrees, minDegrees;
+  vector<OBNodeBase*>::iterator i;
+  vector<OBEdgeBase*>::iterator j,k;
+
+  minDegrees = 360.0;
+
+  for (b = BeginNbrAtom(j); b; b = NextNbrAtom(j))
+    {
+      k = j;
+      for (c = NextNbrAtom(k); c; c = NextNbrAtom(k))
+	{
+	  v1 = b->GetVector() - GetVector();
+	  v2 = c->GetVector() - GetVector();	
+	  degrees = vectorAngle(v1, v2);
+	  if (degrees < minDegrees)
+	    minDegrees = degrees;
+	}
+    }
+  return minDegrees;
+}
+
 unsigned int OBAtom::CountFreeOxygens() const
 {
   unsigned int count = 0;
   OBAtom *atom;
+  OBBond *bond;
   vector<OBEdgeBase*>::iterator i;
 
-  for (atom = ((OBAtom*)this)->BeginNbrAtom(i);atom;atom = ((OBAtom*)this)->NextNbrAtom(i))
-	  if (atom->IsOxygen() && atom->GetHvyValence() == 1)
-		  count++;
+  for (bond = ((OBAtom*)this)->BeginBond(i);bond;bond = ((OBAtom*)this)->NextBond(i))
+    {
+      atom = bond->GetNbrAtom((OBAtom*)this);
+      if (atom->IsOxygen() && atom->GetHvyValence() == 1)
+	count++;
+    }
 
   return(count);
 }
@@ -616,7 +801,7 @@ unsigned int OBAtom::BOSum() const
 unsigned int OBAtom::KBOSum() const
 {
   OBBond *bond;
-  unsigned int bosum = 0;
+  unsigned int bosum;
   vector<OBEdgeBase*>::iterator i;
   
   bosum = GetImplicitValence();
@@ -665,6 +850,23 @@ bool OBAtom::DeleteBond(OBBond *bond)
   return(false);
 }
 
+bool OBAtom::MatchesSMARTS(const char *pattern)
+{
+  OBMol *mol = (OBMol*)((OBAtom*)this)->GetParent();
+  vector<vector<int> > mlist;
+  vector<vector<int> >::iterator l;
+
+  OBSmartsPattern test; test.Init(pattern);
+  if (test.Match(*mol))
+    {
+      mlist = test.GetUMapList();
+      for (l = mlist.begin(); l != mlist.end(); l++)
+	if (GetIdx() == mol->GetAtom((*l)[0])->GetIdx())
+	  return true;
+    }
+  return false;
+}
+
 OBBond *OBAtom::BeginBond(vector<OBEdgeBase*>::iterator &i) 
 {
 	i = _vbond.begin();
@@ -689,14 +891,46 @@ OBAtom *OBAtom::NextNbrAtom(vector<OBEdgeBase*>::iterator &i)
   return((i != _vbond.end()) ?  ((OBBond*) *i)->GetNbrAtom(this):NULL);
 }
 
+double OBAtom::GetDistance(OBAtom *b)
+{
+  return(( this->GetVector() - b->GetVector() ).length());
+}
 
-bool OBAtom::GetNewBondVector(Vector &v,float length)
+double OBAtom::GetDistance(int b)
+{
+  OBMol *mol = (OBMol*)GetParent();
+  return(( this->GetVector() - mol->GetAtom(b)->GetVector() ).length());
+}
+
+double OBAtom::GetAngle(OBAtom *b, OBAtom *c)
+{
+  vector3 v1,v2;
+
+  v1 = this->GetVector() - b->GetVector();
+  v2 = c->GetVector() - b->GetVector();
+
+  return(vectorAngle(v1, v2));
+}
+
+double OBAtom::GetAngle(int b, int c)
+{
+  OBMol *mol = (OBMol*)GetParent();
+  vector3 v1,v2;
+
+  v1 = this->GetVector() - mol->GetAtom(b)->GetVector();
+  v2 = mol->GetAtom(c)->GetVector() - mol->GetAtom(b)->GetVector();
+
+  return(vectorAngle(v1, v2));
+}
+
+bool OBAtom::GetNewBondVector(vector3 &v,double length)
 {
   // ***experimental code***
 
   OBAtom *atom;
   vector<OBEdgeBase*>::iterator i,j;
   v = VZero;
+
   if (GetValence() == 0)
     {
       v = VX;
@@ -707,7 +941,7 @@ bool OBAtom::GetNewBondVector(Vector &v,float length)
 
   if (GetValence() == 1)
     {
-      Vector vtmp,v1,v2;
+      vector3 vtmp,v1,v2;
       atom = BeginNbrAtom(i);
       if (atom)
 	vtmp = GetVector() - atom->GetVector();
@@ -734,18 +968,17 @@ bool OBAtom::GetNewBondVector(Vector &v,float length)
 	  else
 	      v1 = cross(vtmp,v2);
 
-	  Matrix3x3 m;
+	  matrix3x3 m;
 	  m.RotAboutAxisByAngle(v1,60.0);
 	  v = m*vtmp;
 	  v.normalize();
 	}
-
-      if (GetHyb() == 3)
+      else if (GetHyb() == 3)
 	{
 	  v1 = cross(vtmp,VX);
 	  v2 = cross(vtmp,VY);
 	  if (v1.length() < v2.length()) v1 = v2;
-	  Matrix3x3 m;
+	  matrix3x3 m;
 	  m.RotAboutAxisByAngle(v1,70.5);
 	  v = m*vtmp;
 	  v.normalize();
@@ -759,7 +992,7 @@ bool OBAtom::GetNewBondVector(Vector &v,float length)
 
   if (GetValence() == 2)
     {
-      Vector v1,v2,vtmp,vsum,vnorm;
+      vector3 v1,v2,vtmp,vsum,vnorm;
       atom = BeginNbrAtom(i);if (!atom) return(false);
       v1 = GetVector() - atom->GetVector();
       atom = NextNbrAtom(i); if (!atom) return(false);
@@ -768,36 +1001,35 @@ bool OBAtom::GetNewBondVector(Vector &v,float length)
       vsum = v1+v2;
       vsum.normalize();
 
-      if (GetHyb() == 2) v = vsum;
-
-      if (GetHyb() == 3)
+      if (GetHyb() == 2)
+	v = vsum;
+      else if (GetHyb() == 3)
 	{
 	  vnorm = cross(v2,v1);
 	  vnorm.normalize();
 
 #ifndef ONE_OVER_SQRT3
-#define ONE_OVER_SQRT3  0.577350269f
+#define ONE_OVER_SQRT3  0.57735026918962576451
 #endif //SQRT_TWO_THIRDS
 #ifndef SQRT_TWO_THIRDS
-#define SQRT_TWO_THIRDS 0.816496581f
+#define SQRT_TWO_THIRDS 0.81649658092772603272
 #endif //ONE_OVER_SQRT3
 
 	  vsum *= ONE_OVER_SQRT3;
 	  vnorm *= SQRT_TWO_THIRDS;
 
-#undef ONE_OVER_SQRT3
-#undef SQRT_TWO_THIRDS
-
 	  v = vsum + vnorm;
 	}
+
       v *= length;
+
       v += GetVector();
       return(true);
     }
 
   if (GetValence() == 3)
     {
-      Vector vtmp,vsum;
+      vector3 vtmp,vsum;
       OBAtom *atom;
       vector<OBEdgeBase*>::iterator i;
       for (atom = BeginNbrAtom(i);atom;atom = NextNbrAtom(i))
@@ -837,14 +1069,14 @@ bool OBAtom::HtoMethyl()
       return(false);
     }
 
-  float br1,br2;
+  double br1,br2;
   br1 = etab.CorrectedBondRad(6,3);
   br2 = etab.CorrectedBondRad(atom->GetAtomicNum(),atom->GetHyb());
   bond->SetLength(atom,br1+br2);
 
   OBAtom *hatom;
   br2 = etab.CorrectedBondRad(1,0);
-  Vector v;
+  vector3 v;
   
   for (int j = 0;j < 3;j++)
     {
@@ -861,13 +1093,13 @@ bool OBAtom::HtoMethyl()
   return(true);
 }
 
-static void ApplyRotMatToBond(OBMol &mol,Matrix3x3 &m,OBAtom *a1,OBAtom *a2)
+static void ApplyRotMatToBond(OBMol &mol,matrix3x3 &m,OBAtom *a1,OBAtom *a2)
 {
   vector<int> children;
   mol.FindChildren(children,a1->GetIdx(),a2->GetIdx());
   children.push_back(a2->GetIdx());
 
-  Vector v;
+  vector3 v;
   vector<int>::iterator i;
   for (i = children.begin();i != children.end();i++)
     {
@@ -883,6 +1115,7 @@ static void ApplyRotMatToBond(OBMol &mol,Matrix3x3 &m,OBAtom *a1,OBAtom *a2)
 bool OBAtom::SetHybAndGeom(int hyb)
 {
   //if (hyb == GetHyb()) return(true);
+  if (GetAtomicNum() == 1) return(false);
   if (hyb == 0 && GetHvyValence() > 1) return(false);
   if (hyb == 1 && GetHvyValence() > 2) return(false);
   if (hyb == 2 && GetHvyValence() > 3) return(false);
@@ -904,27 +1137,34 @@ bool OBAtom::SetHybAndGeom(int hyb)
   for (j = delatm.begin();j != delatm.end();j++) mol->DeleteAtom(*j);
   mol->DecrementMod();
 
-  float targetAngle;
+  double targetAngle;
   if (hyb == 3)     targetAngle = 109.5;
   else if (hyb == 2) targetAngle = 120.0;
   else if (hyb == 1) targetAngle = 180.0;
   else               targetAngle = 0.0;
+  
+  if (IsInRing())
+    targetAngle = 180.0 - (360.0 / MemberOfRingSize());
 
   //adjust attached acyclic bond lengths
-  float br1,br2;
+  double br1,br2, length;
   br1 = etab.CorrectedBondRad(GetAtomicNum(),hyb);
   for (nbr = BeginNbrAtom(i);nbr;nbr = NextNbrAtom(i))
     if (!(*i)->IsInRing())
     {
       br2 = etab.CorrectedBondRad(nbr->GetAtomicNum(),nbr->GetHyb());
-      ((OBBond*) *i)->SetLength(this,br1+br2);
+      length = br1 + br2;
+      if ((*i)->IsAromatic()) length *= 0.93;
+      else if ((*i)->GetBO() == 2)         length *= 0.91;
+      else if ((*i)->GetBO() == 3)         length *= 0.87;
+      ((OBBond*) *i)->SetLength(this, length);
     }
 
   if (GetValence() > 1)
     {
-      float angle;
-      Matrix3x3 m;
-      Vector v1,v2,v3,v4,n,s;
+      double angle;
+      matrix3x3 m;
+      vector3 v1,v2,v3,v4,n,s;
       OBAtom *r1,*r2,*r3,*a1,*a2,*a3,*a4;
       r1 = r2 = r3 = a1 = a2 = a3 = a4 = NULL;
 
@@ -951,7 +1191,7 @@ bool OBAtom::SetHybAndGeom(int hyb)
 	      v1 = a1->GetVector()-GetVector();v1.normalize();
 	      v2 = a2->GetVector()-GetVector();v2.normalize();
 	      n = cross(v1,v2);
-	      angle = VectorAngle(v1,v2)-targetAngle;
+	      angle = vectorAngle(v1,v2)-targetAngle;
 	      m.RotAboutAxisByAngle(n,-angle);
 	      ApplyRotMatToBond(*mol,m,this,a1);
 	    }
@@ -963,9 +1203,9 @@ bool OBAtom::SetHybAndGeom(int hyb)
 	      v1 = r1->GetVector()-GetVector();v1.normalize();
 	      v2 = r2->GetVector()-GetVector();v2.normalize();
 	      v3 = a1->GetVector()-GetVector();
-	      s = v1+v2; s.normalize(); s *= -1.0f;
+	      s = v1+v2; s.normalize(); s *= -1.0;
 	      n = cross(s,v3);
-	      angle = VectorAngle(s,v3);
+	      angle = vectorAngle(s,v3);
 	      m.RotAboutAxisByAngle(n,angle);
 	      ApplyRotMatToBond(*mol,m,this,a1);
 	    }
@@ -976,7 +1216,7 @@ bool OBAtom::SetHybAndGeom(int hyb)
 		  v1 = a1->GetVector()-GetVector();v1.normalize();
 		  v2 = a2->GetVector()-GetVector();v2.normalize();
 		  n = cross(v1,v2);
-		  angle = VectorAngle(v1,v2)-targetAngle;
+		  angle = vectorAngle(v1,v2)-targetAngle;
 		  m.RotAboutAxisByAngle(n,-angle);
 		  ApplyRotMatToBond(*mol,m,this,a1);
 		}
@@ -985,9 +1225,9 @@ bool OBAtom::SetHybAndGeom(int hyb)
 		  v1 = a1->GetVector()-GetVector();v1.normalize();
 		  v2 = a2->GetVector()-GetVector();v2.normalize();
 		  v3 = a3->GetVector()-GetVector();
-		  s = v1+v2; s.normalize(); s *= -1.0f;
+		  s = v1+v2; s.normalize(); s *= -1.0;
 		  n = cross(s,v3);
-		  angle = VectorAngle(s,v3);
+		  angle = vectorAngle(s,v3);
 		  m.RotAboutAxisByAngle(n,angle);
 		  ApplyRotMatToBond(*mol,m,this,a3);
 		}
@@ -1001,9 +1241,9 @@ bool OBAtom::SetHybAndGeom(int hyb)
 	      v2 = r2->GetVector()-GetVector();v2.normalize();
 	      v3 = r3->GetVector()-GetVector();v3.normalize();
 	      v4 = a1->GetVector()-GetVector();
-	      s = v1 + v2 + v3; s *= -1.0f; s.normalize();
+	      s = v1 + v2 + v3; s *= -1.0; s.normalize();
 	      n = cross(s,v4);
-	      angle = VectorAngle(s,v4);
+	      angle = vectorAngle(s,v4);
 	      m.RotAboutAxisByAngle(n,angle);
 	      ApplyRotMatToBond(*mol,m,this,a1);
 	    }
@@ -1012,13 +1252,13 @@ bool OBAtom::SetHybAndGeom(int hyb)
 	      v1 = r1->GetVector()-GetVector();v1.normalize();
 	      v2 = r2->GetVector()-GetVector();v2.normalize();
 	      v3 = a1->GetVector()-GetVector();
-	      s = v1+v2; s *= -1.0f; s.normalize(); 
+	      s = v1+v2; s *= -1.0; s.normalize(); 
 	      n = cross(v1,v2);      n.normalize();
-	      s *= 0.577350269f; //1/sqrt(3)
-	      n *= 0.816496581f; //sqrt(2/3)
+	      s *= ONE_OVER_SQRT3; //1/sqrt(3)
+	      n *= SQRT_TWO_THIRDS; //sqrt(2/3)
 	      s += n; s.normalize();
 	      n = cross(s,v3);
-	      angle = VectorAngle(s,v3);
+	      angle = vectorAngle(s,v3);
 	      m.RotAboutAxisByAngle(n,angle);
 	      ApplyRotMatToBond(*mol,m,this,a1);
 
@@ -1028,9 +1268,9 @@ bool OBAtom::SetHybAndGeom(int hyb)
 		  v2 = r2->GetVector()-GetVector();v2.normalize();
 		  v3 = a1->GetVector()-GetVector();v3.normalize();
 		  v4 = a2->GetVector()-GetVector();
-		  s = v1 + v2 + v3; s *= -1.0f; s.normalize();
+		  s = v1 + v2 + v3; s *= -1.0; s.normalize();
 		  n = cross(s,v4);
-		  angle = VectorAngle(s,v4);
+		  angle = vectorAngle(s,v4);
 		  m.RotAboutAxisByAngle(n,angle);
 		  ApplyRotMatToBond(*mol,m,this,a2);
 		}
@@ -1042,7 +1282,7 @@ bool OBAtom::SetHybAndGeom(int hyb)
 		  v1 = a1->GetVector()-GetVector();v1.normalize();
 		  v2 = a2->GetVector()-GetVector();v2.normalize();
 		  n = cross(v1,v2);
-		  angle = VectorAngle(v1,v2)-targetAngle;
+		  angle = vectorAngle(v1,v2)-targetAngle;
 		  m.RotAboutAxisByAngle(n,-angle);
 		  ApplyRotMatToBond(*mol,m,this,a1);
 		}
@@ -1051,13 +1291,13 @@ bool OBAtom::SetHybAndGeom(int hyb)
 		  v1 = a1->GetVector()-GetVector();v1.normalize();
 		  v2 = a2->GetVector()-GetVector();v2.normalize();
 		  v3 = a3->GetVector()-GetVector();
-		  s = v1+v2; s *= -1.0f; s.normalize(); 
+		  s = v1+v2; s *= -1.0; s.normalize(); 
 		  n = cross(v1,v2);      n.normalize();
-		  s *= 0.577350269f; //1/sqrt(3)
-		  n *= 0.816496581f; //sqrt(2/3)
+		  s *= ONE_OVER_SQRT3; //1/sqrt(3)
+		  n *= SQRT_TWO_THIRDS; //sqrt(2/3)
 		  s += n; s.normalize();
 		  n = cross(s,v3);
-		  angle = VectorAngle(s,v3);
+		  angle = vectorAngle(s,v3);
 		  m.RotAboutAxisByAngle(n,angle);
 		  ApplyRotMatToBond(*mol,m,this,a3);
 		}
@@ -1101,9 +1341,9 @@ bool OBAtom::SetHybAndGeom(int hyb)
   if (hcount)
     {
       int k;
-      Vector v;
+      vector3 v;
       OBAtom *atom;
-      float brsum = etab.CorrectedBondRad(1,0)+
+      double brsum = etab.CorrectedBondRad(1,0)+
 	etab.CorrectedBondRad(GetAtomicNum(),GetHyb());
       SetHyb(hyb);
 
@@ -1134,4 +1374,123 @@ OBBond *OBAtom::GetBond(OBAtom *nbr)
   return NULL;
 }
 
+// OBGenericData methods
+bool OBAtom::HasData(string &s)
+     //returns true if the generic attribute/value pair exists
+{
+  if (_vdata.empty()) return(false);
+
+    vector<OBGenericData*>::iterator i;
+
+    for (i = _vdata.begin();i != _vdata.end();i++)
+        if ((*i)->GetAttribute() == s)
+            return(true);
+    
+    return(false);
+}
+
+bool OBAtom::HasData(const char *s)
+     //returns true if the generic attribute/value pair exists
+{
+  if (_vdata.empty()) return(false);
+
+    vector<OBGenericData*>::iterator i;
+
+    for (i = _vdata.begin();i != _vdata.end();i++)
+        if ((*i)->GetAttribute() == s)
+            return(true);
+    
+    return(false);
+}
+
+bool OBAtom::HasData(obDataType dt)
+     //returns true if the generic attribute/value pair exists
+{
+  if (_vdata.empty()) return(false);
+
+    vector<OBGenericData*>::iterator i;
+
+    for (i = _vdata.begin();i != _vdata.end();i++)
+        if ((*i)->GetDataType() == dt)
+            return(true);
+    
+    return(false);
+}
+
+OBGenericData *OBAtom::GetData(string &s)
+     //returns the value given an attribute
+{
+    vector<OBGenericData*>::iterator i;
+
+    for (i = _vdata.begin();i != _vdata.end();i++)
+                if ((*i)->GetAttribute() == s)
+            return(*i);
+
+    return(NULL);
+}
+
+OBGenericData *OBAtom::GetData(const char *s)
+     //returns the value given an attribute
+{
+    vector<OBGenericData*>::iterator i;
+
+    for (i = _vdata.begin();i != _vdata.end();i++)
+                if ((*i)->GetAttribute() == s)
+            return(*i);
+
+    return(NULL);
+}
+
+OBGenericData *OBAtom::GetData(obDataType dt)
+{
+    vector<OBGenericData*>::iterator i;
+    for (i = _vdata.begin();i != _vdata.end();i++)
+        if ((*i)->GetDataType() == dt)
+            return(*i);
+    return(NULL);
+}
+
+void OBAtom::DeleteData(obDataType dt)
+{
+  vector<OBGenericData*> vdata;
+  vector<OBGenericData*>::iterator i;
+    for (i = _vdata.begin();i != _vdata.end();i++)
+        if ((*i)->GetDataType() == dt) delete *i;
+        else vdata.push_back(*i);
+  _vdata = vdata;
+}
+
+void OBAtom::DeleteData(vector<OBGenericData*> &vg)
+{
+  vector<OBGenericData*> vdata;
+  vector<OBGenericData*>::iterator i,j;
+
+  bool del;
+  for (i = _vdata.begin();i != _vdata.end();i++)
+  {
+          del = false;
+          for (j = vg.begin();j != vg.end();j++)
+                  if (*i == *j)
+                  {
+                          del = true;
+                          break;
+                  }
+           if (del) delete *i;
+           else     vdata.push_back(*i);
+  }
+  _vdata = vdata;
+}
+
+void OBAtom::DeleteData(OBGenericData *gd)
+{
+  vector<OBGenericData*>::iterator i;
+  for (i = _vdata.begin();i != _vdata.end();i++)
+          if (*i == gd)
+          {
+                delete *i;
+                _vdata.erase(i);
+          }
+
+}
+
 }
