@@ -1,5 +1,11 @@
 /**********************************************************************
+base.cpp - Base classes to build a graph
+
 Copyright (C) 1998-2001 by OpenEye Scientific Software, Inc.
+Some portions Copyright (c) 2001-2003 by Geoffrey R. Hutchison
+
+This file is part of the Open Babel project.
+For more information, see <http://openbabel.sourceforge.net/>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -11,13 +17,25 @@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 ***********************************************************************/
 
+#include "babelconfig.h"
+
 #include "base.h"
 
+#if HAVE_IOSTREAM
 #include <iostream>
-#include <fstream>
+#elif HAVE_IOSTREAM_H
+#include <iostream.h>
+#endif
 
-namespace OpenBabel
-{
+#if HAVE_FSTREAM
+#include <fstream>
+#elif HAVE_FSTREAM_H
+#include <fstream.h>
+#endif
+
+using namespace std;
+
+namespace OpenBabel {
 
 bool OBGraphBase::Match(OBGraphBase &g,bool singleMatch)
 {
@@ -193,4 +211,62 @@ bool OBGraphBase::SetVisitLock(bool v)
 	return(true);
 }
 
-}
+/*! \mainpage
+  \section intro Introduction
+
+  (portions adapted from the OELib primer, written by Matt Stahl, Open
+   Eye Scientific Software, Inc.)
+
+It is fair to say that Open Babel (and before it, OELib) is a direct
+result of the original Babel. Programmers have long known that
+application development is facilitated by building software on top of
+libraries rich in functionality. Babel was the first experience for
+Matt Stahl in designing a molecule library. In addition to developing
+Babel, Pat Walters and Matt developed `OBabel' at Vertex
+Pharmaceuticals. OBabel was the first attempt at developing an 
+object oriented molecule library. Although OBabel was a successful
+project, Matt's departure from Vertex Pharmaceuticals provided a great
+opportunity to include much of what I had learned in previous projects
+into a new molecule class library. OELib was then designed to be flexible,
+extensible, portable, and efficient class library for processing small
+molecules.
+
+OELib was released under the GNU General Public License (GPL) by Matt Stahl
+and Open Eye Scientific Software, Inc. to take advantage of many of
+the "great minds writing chemical software." Eventually Open Eye
+decided to write a next-generation class library as proprietary
+software. The result was that Open Babel took up where OELib left off,
+using the existing GPL version of OELib, and has continued to evolve
+and improve.
+
+There are several advantages to having the source code to Open Babel
+available. First, development time can be shortened by basing projects
+on Open Babel. Many chemical and molecular concepts and code are
+already implemented. The fewer people who have to reinvent the wheel
+(or the function), the better. Second, as free software, hopefully 
+other programmers will contribute to the project.
+
+Thanks to all who have helped with Babel, OBabel, OELib and Open Babel.
+The list is long and growing.
+
+  \section pointers Key Modules
+
+The heart of Open Babel lies in the \link OpenBabel::OBMol OBMol\endlink, 
+\link OpenBabel::OBAtom OBAtom\endlink, and 
+\link OpenBabel::OBBond OBBond\endlink classes,
+which handle operations on atoms, bonds and molecules. Newcomers should 
+start with looking at the \link OpenBabel::OBMol OBMol\endlink class, 
+designed to store the basic information
+in a molecule and to perceive information about a molecule.
+
+
+transformations and automatic perception of properties is performed in a 
+"lazy" manner. That is, until you call for partial atomic charges, no 
+charges are calculated. This ensures faster transformations of chemical data
+-- properties that are not needed for your code will typically not be 
+calculated.
+
+*/
+
+} // namespace OpenBabel
+
