@@ -1,5 +1,6 @@
 /**********************************************************************
 Copyright (C) 1998-2001 by OpenEye Scientific Software, Inc.
+Some portions Copyright (c) 2001-2003 by Geoffrey R. Hutchison
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -17,12 +18,23 @@ GNU General Public License for more details.
 
 //#define KEKULE
 
+using namespace std;
+
 namespace OpenBabel {
 
-bool WriteSmiles(ostream &ofs,OBMol &mol,char *title)
+bool WriteSmiles(ostream &ofs,OBMol &mol,const char *title)
 {
   char buffer[BUFF_SIZE],tmp[BUFF_SIZE];
 
+  // This is a hack to prevent recursion problems.
+  //  we still need to fix the underlying problem -GRH
+  if (mol.NumAtoms() > 1000)
+    {
+      ThrowError("SMILES Conversion failed: Molecule is too large to convert.");
+      cerr << "  Molecule size: " << mol.NumAtoms() << " atoms " << endl;
+      return(false);
+    }
+
   OBMol2Smi m2s;
 
   m2s.Init();
@@ -225,8 +237,8 @@ void OBMol2Smi::FindClosureBonds(OBMol &mol)
       a1 = a2 = NULL;
 
       for (k = _storder.begin();k != _storder.end();k++)
-	if (bond->GetBeginAtomIdx() == *k || 
-	    bond->GetEndAtomIdx() == *k)
+	if (bond->GetBeginAtomIdx() == static_cast<unsigned int>(*k) || 
+	    bond->GetEndAtomIdx() == static_cast<unsigned int>(*k))
 	  if (!a1) a1 = mol.GetAtom(*k);
 	  else if (!a2) 
 	    {
@@ -236,7 +248,7 @@ void OBMol2Smi::FindClosureBonds(OBMol &mol)
 	    }
 
       for (k = _storder.begin();k != _storder.end();k++)
-		if (a1->GetIdx() == *k)
+		if (a1->GetIdx() == static_cast<unsigned int>(*k))
 		{
 			k++;
 			if (k != _storder.end()) _storder.insert(k,a2->GetIdx());
@@ -393,6 +405,18 @@ bool OBMol2Smi::GetSmilesElement(OBSmiNode *node,char *element)
 
   if (atom->GetFormalCharge() != 0) //bracket charged elements
     bracketElement = true;
+	
+	//CM begin 18 Sept 2003
+
+	//This outputs form [CH3][CH3] rather than CC if -h option has been specified
+	if (((OBMol*)atom->GetParent())->HasHydrogensAdded())
+    bracketElement = true;
+
+	//This outputs explicit H form anyway for radicals
+	if (atom->GetSpinMultiplicity())
+    bracketElement = true;
+
+	//CM end
 
   if (!bracketElement)
     {
@@ -507,7 +531,7 @@ bool OBMol2Smi::GetSmilesElement(OBSmiNode *node,char *element)
 bool OBMol2Smi::GetChiralStereo(OBSmiNode *node,char *stereo)
 {
   bool is2D=false;
-  float torsion;
+  double torsion;
   OBAtom *a,*b,*c,*d,hydrogen;
 
   b = node->GetAtom();
@@ -526,7 +550,7 @@ bool OBMol2Smi::GetChiralStereo(OBSmiNode *node,char *stereo)
   //give peudo Z coords if mol is 2D
   if (!mol->Has3D())
     {
-      Vector v,vz(0.0,0.0,1.0);
+      vector3 v,vz(0.0,0.0,1.0);
       is2D = true;
 	  OBAtom *nbr;
       OBBond *bond;
@@ -572,7 +596,7 @@ bool OBMol2Smi::GetChiralStereo(OBSmiNode *node,char *stereo)
 	}
       else  //implicit hydrogen
 	{
-	  Vector v;
+	  vector3 v;
 	  b->GetNewBondVector(v,1.0);
 	  hydrogen.SetVector(v);
 	  c = &hydrogen;
@@ -615,7 +639,7 @@ bool OBMol2Smi::GetChiralStereo(OBSmiNode *node,char *stereo)
   //re-zero psuedo-coords
   if (is2D)
     {
-      Vector v;
+      vector3 v;
       OBAtom *atom;
       vector<OBNodeBase*>::iterator k;
       for (atom = mol->BeginAtom(k);atom;atom = mol->NextAtom(k))
@@ -634,6 +658,15 @@ bool WriteFixFile(ostream &ofs,OBMol &mol)
   char buffer[BUFF_SIZE];
   OBMol2Smi m2s;
 
+  // This is a hack to prevent recursion problems.
+  //  we still need to fix the underlying problem -GRH
+  if (mol.NumAtoms() > 1000)
+    {
+      ThrowError("SMILES Conversion failed: Molecule is too large to convert.");
+      cerr << "  Molecule size: " << mol.NumAtoms() << " atoms " << endl;
+      return(false);
+    }
+
   m2s.Init();
   //m2s.AssignCisTrans(mol);
   m2s.CorrectAromaticAmineCharge(mol);
